/**
 * Copyright 2015 Confluent Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 **/

package io.confluent.connect.jdbc;

import org.apache.kafka.connect.connector.ConnectorContext;
import org.easymock.EasyMock;
import org.easymock.IAnswer;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;
import org.junit.runner.RunWith;
import org.powermock.api.easymock.PowerMock;
import org.powermock.api.easymock.annotation.Mock;
import org.powermock.core.classloader.annotations.PowerMockIgnore;
import org.powermock.core.classloader.annotations.PrepareForTest;
import org.powermock.modules.junit4.PowerMockRunner;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;

import static org.junit.Assert.assertEquals;

@RunWith(PowerMockRunner.class)
@PrepareForTest({TableMonitorThread.class, JdbcUtils.class})
@PowerMockIgnore("javax.management.*")
public class TableMonitorThreadTest {
  private static final long POLL_INTERVAL = 100;

  private static final List<String> FIRST_TOPIC_LIST = Arrays.asList("foo");
  private static final List<String> SECOND_TOPIC_LIST = Arrays.asList("foo", "bar");
  private static final List<String> THIRD_TOPIC_LIST = Arrays.asList("foo", "bar", "baz");

  private EmbeddedDerby db;
  private Connection dbConn;
  private TableMonitorThread tableMonitorThread;
  @Mock private ConnectorContext context;

  @Before
  public void setUp() throws SQLException {
    db = new EmbeddedDerby();
    dbConn = DriverManager.getConnection(db.getUrl());

    PowerMock.mockStatic(JdbcUtils.class);
  }

  @After
  public void tearDown() throws Exception {
    db.close();
    db.dropDatabase();
  }

  @Test
  public void testSingleLookup() throws Exception {
    tableMonitorThread = new TableMonitorThread(dbConn, context, POLL_INTERVAL, null, null);

    EasyMock.expect(JdbcUtils.getTables(dbConn)).andAnswer(new IAnswer<List<String>>() {
      @Override
      public List<String> answer() throws Throwable {
        tableMonitorThread.shutdown();
        return FIRST_TOPIC_LIST;
      }
    });

    PowerMock.replayAll();

    tableMonitorThread.start();
    tableMonitorThread.join();
    assertEquals(FIRST_TOPIC_LIST, tableMonitorThread.tables());

    PowerMock.verifyAll();
  }

  @Test
  public void testWhitelist() throws Exception {
    tableMonitorThread = new TableMonitorThread(dbConn, context, POLL_INTERVAL,
                                                new HashSet<>(Arrays.asList("foo", "bar")), null);

    EasyMock.expect(JdbcUtils.getTables(dbConn)).andAnswer(new IAnswer<List<String>>() {
      @Override
      public List<String> answer() throws Throwable {
        tableMonitorThread.shutdown();
        return THIRD_TOPIC_LIST;
      }
    });

    PowerMock.replayAll();

    tableMonitorThread.start();
    tableMonitorThread.join();
    assertEquals(Arrays.asList("foo", "bar"), tableMonitorThread.tables());

    PowerMock.verifyAll();
  }

  @Test
  public void testBlacklist() throws Exception {
    tableMonitorThread = new TableMonitorThread(dbConn, context, POLL_INTERVAL,
                                                null, new HashSet<>(Arrays.asList("bar", "baz")));

    EasyMock.expect(JdbcUtils.getTables(dbConn)).andAnswer(new IAnswer<List<String>>() {
      @Override
      public List<String> answer() throws Throwable {
        tableMonitorThread.shutdown();
        return THIRD_TOPIC_LIST;
      }
    });

    PowerMock.replayAll();

    tableMonitorThread.start();
    tableMonitorThread.join();
    assertEquals(Arrays.asList("foo"), tableMonitorThread.tables());

    PowerMock.verifyAll();
  }

  @Test
  public void testReconfigOnUpdate() throws Exception {
    tableMonitorThread = new TableMonitorThread(dbConn, context, POLL_INTERVAL, null, null);

    EasyMock.expect(JdbcUtils.getTables(dbConn)).andReturn(FIRST_TOPIC_LIST);
    // Returning same list should not change results
    EasyMock.expect(JdbcUtils.getTables(dbConn)).andAnswer(new IAnswer<List<String>>() {
      @Override
      public List<String> answer() throws Throwable {
        assertEquals(FIRST_TOPIC_LIST, tableMonitorThread.tables());
        return FIRST_TOPIC_LIST;
      }
    });
    // Changing the result should trigger a task reconfiguration
    EasyMock.expect(JdbcUtils.getTables(dbConn)).andReturn(SECOND_TOPIC_LIST);
    context.requestTaskReconfiguration();
    PowerMock.expectLastCall();
    // Changing again should result in another update
    EasyMock.expect(JdbcUtils.getTables(dbConn)).andAnswer(new IAnswer<List<String>>() {
      @Override
      public List<String> answer() throws Throwable {
        assertEquals(SECOND_TOPIC_LIST, tableMonitorThread.tables());
        tableMonitorThread.shutdown();
        return FIRST_TOPIC_LIST;
      }
    });
    context.requestTaskReconfiguration();
    PowerMock.expectLastCall();

    PowerMock.replayAll();

    tableMonitorThread.start();
    tableMonitorThread.join();
    assertEquals(FIRST_TOPIC_LIST, tableMonitorThread.tables());

    PowerMock.verifyAll();
  }
}
